/***************************************************************************/
/*                                                                         */
/*  Hoehe.c                                                                */
/*             Luftdruck-/Hoehenmesser  und Temperatur mit LCD Anzeige     */
/*             PIC16F877  4Mhz                                             */
/*                                                                         */
/*      Mode : HD44780 type LCD displays                                   */
/*                                                                         */
/* Interface :     SEL = Port B 4                                          */
/*                  RS = Port B 5                                          */
/*              Data_7 = Port B 3                                          */
/*              Data_6 = Port B 2                                          */
/*              Data_5 = Port B 1                                          */
/*              Data_4 = Port B 0                                          */
/*          MPXS4100A  = Port A 0  // Luftdrucksensor                      */
/*               VR+   = Port A 3  // +5V Referenzspannung                 */
/*    Taster1 (ROT)    = Port C 0  // Taster1 Input (Mode,Set)             */
/*    Taster2 (SCHWARZ)= Port C 1  // Taster2 Input (+)                    */
/*    Taster3 (SCHWARZ)= Port C 2  // Taster3 Input (-)                    */
/*          TCN75  SDA = Port C 4  // Temperatursensor per I2C Bus         */
/*          TCN75  SCL = Port C 3  //       "                              */
/*                                                                         */
/* (c) Matthias Stehr 20.05.2001                                           */
/*                                                                         */
/*                                                                         */
/***************************************************************************/
#include <16F877.h>
#use i2c(MASTER, SDA=PIN_C4, SCL=PIN_C3)

//#define DEBUG
#define TARGET

#ifdef DEBUG
#use delay(clock=20000000)         // Quarz auf Entwicklungsboard
#else
#use delay(clock=4000000)          // Quarz auf realer Hardware
#endif // DEBUG

#use rs232(baud=9600, xmit=PIN_C6, rcv=PIN_C7)

#include <math.h>
#ifdef DEBUG
#include <lcd_d.c>    // LCD an Port D
#else
#include <lcd_b.c>    // LCD an Port B
#endif // DEBUG


void hoehe(void);
void setp0(void);
void writeEEp0(void);
void readEEp0(void);
void messen(void);
void eichen(void);
void messzeit(void);
void readTemperatur(void);
void readEEProm(void);
byte read_ext_eeprom(long int address);
void write_ext_eeprom(long int address, byte data);
void initEEProm(void);


#define PortBConfig  0b11111111   /* Configure port B  0 = output       */
#define PortCConfig  0b10000111   /* Configure port C  0 = output       */
#define EE_P0        0x00         // Addr. im EE-Prom  + 3 Byte !!!!
#define EE_OFFSET    0x04
#define EE_DT        0x05         // Mess-Zeitintervall in Sekunden


#bit ADFM = 0x9F.7     // A/D Result format
float dh, druck, p0;
char offset, temperaturMSB, temperaturLSB, time;
short store;
long index;



/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void main()
{
   signed long h;

   #ifdef DEBUG
   set_tris_b( PortBConfig );
   #endif
   set_tris_c( PortCConfig );
   #ifdef DEBUG
   setup_adc_ports(RA0_RA1_RA3_ANALOG);
   #else
   setup_adc_ports(RA0_RA1_ANALOG_RA3_REF);
   #endif //DEBUG
   setup_adc(ADC_CLOCK_INTERNAL);
   ADFM = TRUE;
   set_adc_channel(0);
   store = FALSE;
   lcd_init();
   readEEp0();
   offset = read_eeprom(EE_OFFSET);
   time = read_eeprom(EE_DT);
   printf(lcd_putc,"\fDesigned by\n(c) M.Stehr 2001");
   delay_ms(2000);
   while(1)
      {
      output_high(PIN_C5);
      messen();
      hoehe();
      readTemperatur();
      lcd_gotoxy(1,1);
      printf(lcd_putc,"%4.0fhPa p0:%4.0f  \n",druck, p0)  ;
      printf(lcd_putc,"dh:%3.0fm", dh);
      if(store)
         printf(lcd_putc,"*");
      printf(lcd_putc,"%D.%U",temperaturMSB, temperaturLSB) ;
      lcd_putc(0xDF); lcd_putc('C');
      printf(lcd_putc,"     ");
      output_low(PIN_C5);
      if(store)
         {
         h = dh;
         write_ext_eeprom(index,h>>8);       // MSB
         index++;
         write_ext_eeprom(index,h);          // LSB
         index++;
         }
      if(index == 2048)
         store = FALSE;
      setp0();
      if(kbhit())
         {
         getc();
         readEEProm();
         }
      }
}


/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void messen(void)
{
   float m;
   char i;
   long messwert,x;

   x = 0;
   for(i= 0; i<16; i++)
      {
      messwert = read_adc();
      x+= messwert;
      }
   messwert = x/16;

   x = 0;
   for(i= 0; i<16; i++)
      {
      messwert = read_adc();
      x+= messwert;
      }
   x = x/16;
   messwert += x;
   messwert /= 2;
   messwert += offset;
   m        = 0.897571 * (float)messwert;  // messwert * Steigung
   druck    = 145.25 + m;                  // + Konstante
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void hoehe(void)
{
   float b;

   b = p0/druck;
   b = log(b);
   dh = b*8000.0;
}


/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void setp0(void)
{
   long i;
   char x;

   for(x= 0; x < time; x++)
   {

   #ifdef DEBUG
   for(i= 0; i<1000;i++)
      {
      if( !input(PIN_B1) )
         {
         delay_ms(10);
         if( !input(PIN_B1) )
            {
            printf(lcd_putc,"\f<Eichen>");
            while( !input(PIN_B1) )
               ;
            eichen();
            messzeit();
            goto ENDE;
            }
         }
      if( !input(PIN_B2) )
         {
         delay_ms(10);
         if( !input(PIN_B2) )
            {
            printf(lcd_putc,"\f");
            while( !input(PIN_B2) )
               ;
            initEEProm();
            store = TRUE;
            index = 0;
            }
         }
      if( !input(PIN_B3) )
         {
         delay_ms(10);
         if( !input(PIN_B3) )
            {
            printf(lcd_putc,"\f");
            while( !input(PIN_B3) )
               ;
            p0 = druck;
            writeEEp0();
            }
         }
      delay_ms(1);
      }


   #else
   for(i=0; i<1000;i++)
      {
      if( !input(PIN_C0) )
         {
         delay_ms(10);
         if( !input(PIN_C0) )
            {
            printf(lcd_putc,"\f<Eichen>");
            while( !input(PIN_C0) )
               ;
            eichen();
            messzeit();
            goto ENDE;
            }
         }
      if( !input(PIN_C1) )
         {
         delay_ms(10);
         if( !input(PIN_C1) )
            {
            printf(lcd_putc,"\f");
            while( !input(PIN_C1) )
               ;
            initEEProm();
            store = TRUE;
            index = 0;
            }
         }
      if( !input(PIN_C2) )
         {
         delay_ms(10);
         if( !input(PIN_C2) )
            {
            printf(lcd_putc,"\f");
            while( !input(PIN_C2) )
               ;
            p0 = druck;
            writeEEp0();
            }
         }
      delay_ms(1);
      }
   #endif // DEBUG
   }
   ENDE:;
}



/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void eichen(void)
{
   #ifdef DEBUG
   while(input(PIN_B1))
      {
      messen();
      printf(lcd_putc,"\n%4.0fhPa cor:%u  ",druck, offset)  ;
      if( !input(PIN_B2) )
         {
         delay_ms(10);
         if( !input(PIN_B2) )
            {
            while( !input(PIN_B2) )
               ;
            offset++;
            }
         }
      if( !input(PIN_B3) )
         {
         delay_ms(10);
         if( !input(PIN_B3) )
            {
            while( !input(PIN_B3) )
               ;
            offset--;
            }
         }
      }
   while( !input(PIN_B1) )
      ;
   #else
   while(input(PIN_C0))
      {
      messen();
      printf(lcd_putc,"\n%4.0fhPa cor:%u  ",druck, offset)  ;
      if( !input(PIN_C1) )
         {
         delay_ms(10);
         if( !input(PIN_C1) )
            {
            while( !input(PIN_C1) )
               ;
            offset++;
            }
         }
      if( !input(PIN_C2) )
         {
         delay_ms(10);
         if( !input(PIN_C2) )
            {
            while( !input(PIN_C2) )
               ;
            offset--;
            }
         }
      }
   while( !input(PIN_C0) )
      ;
   #endif
   write_eeprom(EE_OFFSET, offset);
}


/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void messzeit(void)
{
   printf(lcd_putc,"\fMessintervall");
   #ifdef DEBUG
   while(input(PIN_B1))
      {
      printf(lcd_putc,"\n%u sec   ",time);
      if( !input(PIN_B2) )
         {
         delay_ms(10);
         if( !input(PIN_B2) )
            {
            while( !input(PIN_B2) )
               ;
            time++;
            if(time == 0)
               time = 1;
            }
         }
      if( !input(PIN_B3) )
         {
         delay_ms(10);
         if( !input(PIN_B3) )
            {
            while( !input(PIN_B3) )
               ;
            time--;
            if(time == 0)
               time = 255;
            }
         }
      }
   while( !input(PIN_B1) )
      ;
   #else
   while(input(PIN_C0))
      {
      printf(lcd_putc,"\n%u sec   ",time);
      if( !input(PIN_C1) )
         {
         delay_ms(10);
         if( !input(PIN_C1) )
            {
            while( !input(PIN_C1) )
               ;
            time++;
            if(time == 0)
               time = 1;
            }
         }
      if( !input(PIN_C2) )
         {
         delay_ms(10);
         if( !input(PIN_C2) )
            {
            while( !input(PIN_C2) )
               ;
            time--;
            if(time == 0)
               time = 255;
            }
         }
      }
   while( !input(PIN_C0) )
      ;
   #endif
   write_eeprom(EE_DT, time);
}



/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void readTemperatur(void)
{
   i2c_Start();                // Set a START condition I2C Bus
   i2c_Write( 0b10010000 );    // Address and Write Flag
   i2c_Write( 0b00000000 );    // Pointer Byte  TEMP Register selection
   i2c_Start();
   i2c_Write( 0b10010001 );    // Address and Read Flag
   temperaturMSB = i2c_Read(); // Read High Byte Teperature
   temperaturLSB = i2c_Read(0);// Read Low Byte Teperature send no ACK
   i2c_stop();
   if((temperaturLSB & 0x80) == 0x80)
      temperaturLSB = 5;
   else
      temperaturLSB = 0;
}




/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void writeEEp0(void)
{
   int i;

   for (i = 0; i < 4; i++)
     write_eeprom(i, *(&p0 + i) ) ;
}



/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void readEEp0(void)
{
   int i;

   for (i = 0; i < 4; i++)
     *(&p0 + i) = read_eeprom(i);
}


/**************************************************************************/
/*                                                                        */
/**************************************************************************/
byte read_ext_eeprom(long int address) {
   byte data;

   i2c_start();
   i2c_write((0xa0|(byte)(address>>7))&0xfe);
   i2c_write(address);
   i2c_start();
   i2c_write((0xa0|(byte)(address>>7))|1);
   data=i2c_read(0);
   i2c_stop();
   return(data);
}



/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void write_ext_eeprom(long int address, byte data) {

   i2c_start();
   i2c_write((0xa0|(byte)(address>>7))&0xfe);
   i2c_write(address);
   i2c_write(data);
   i2c_stop();
   delay_ms(11);
}



/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void readEEProm(void)
{
  long int i;
  signed long h;

  i=0;
  while(TRUE)
   {
   h = read_ext_eeprom(i);
   i++;
   h = h << 8;
   h = h | read_ext_eeprom(i);
   //printf("%lu:0x%2x\n\r",i, read_ext_eeprom(i));
   printf("%ld,\r", h);
   if(i == 2047)
      {
      printf("\n");
      break;
      }
   i++;
   }

}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void initEEProm(void)
{
  long int i;

  printf(lcd_putc,"\fInit EEProm...\ndauert 20sec");
  i=0;
  while(TRUE)
   {
   write_ext_eeprom(i,0x00);
   if(i == 2047)
      break;
   i++;
   }
}

